from typing import Optional

from numpy import abs as _abs
from numpy import float64
from numpy import log as _log
from pandas import DataFrame, Series
from pandas import read_csv as _read_csv

from tompy.matrix import (
    matrix_describe,
    matrix_describe_meta,
    matrix_lr2crdd,
    matrix_lr2sr,
    matrix_lr_metrics,
    matrix_lr_sr_metrics,
    matrix_lr_sr_metrics_meta,
    matrix_portfolio_price,
    matrix_price2lr,
    matrix_price2sr,
    matrix_price_metrics,
    matrix_rolling_max,
    matrix_rolling_mean,
    matrix_rolling_min,
    matrix_rolling_std,
    matrix_rolling_sum,
    matrix_rolling_var,
    matrix_shift,
    matrix_sr2lr,
    matrix_sr_metrics,
)
from tompy.series import (
    series_rolling_corr,
    series_rolling_mean,
    series_rolling_std,
    series_shift,
)


def dataframe_drop_weekends(df: DataFrame, /) -> DataFrame:
    return df[df.index.dayofweek < 5]


def dataframe_read_csv(
    fpath: str, /, *, index_col: Optional[int] = 0, parse_dates: bool = True
) -> DataFrame:
    return _read_csv(fpath, index_col=index_col, parse_dates=parse_dates)


def dataframe_write_csv(df: DataFrame, /, fpath: str) -> None:
    df.to_csv(fpath)


def dataframe_describe(df: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=matrix_describe(df.values),
        index=matrix_describe_meta().keys(),
        columns=df.columns,
        dtype=float64,
    )


def dataframe_shift(df: DataFrame, /, lag: int) -> DataFrame:
    if lag == 0:
        return df
    return DataFrame(
        data=matrix_shift(df.values, lag),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_sum(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_sum(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_mean(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_mean(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_std(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_std(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_var(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_var(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_min(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_min(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_max(df: DataFrame, /, window: int) -> DataFrame:
    return DataFrame(
        data=matrix_rolling_max(df.values, window),
        index=df.index,
        columns=df.columns,
        dtype=float64,
    )


def dataframe_rolling_corr(
    df: DataFrame, /, window: int, *, cidx: int = 0
) -> DataFrame:
    s = df[df.columns[cidx]]
    df2 = DataFrame(index=df.index, columns=df.columns, dtype=float64)
    for col in df:
        df2[col] = series_rolling_corr(s, df[col], window)
    return df2


def dataframe_rolling_ema(df: DataFrame, /, window: int) -> DataFrame:
    return df.ewm(span=window, adjust=False).mean()


def dataframe_price2sr(price: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=matrix_price2sr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=float64,
    )


def dataframe_price2lr(price: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=matrix_price2lr(price.values),
        index=price.index,
        columns=price.columns,
        dtype=float64,
    )


def dataframe_price_metrics(
    price: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=matrix_price_metrics(
            price.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=matrix_lr_sr_metrics_meta(mode).keys(),
        columns=price.columns,
        dtype=float64,
    )


def dataframe_sr2lr(sr: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=matrix_sr2lr(sr.values),
        index=sr.index,
        columns=sr.columns,
        dtype=float64,
    )


def dataframe_sr_metrics(
    sr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=matrix_sr_metrics(
            sr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=matrix_lr_sr_metrics_meta(mode).keys(),
        columns=sr.columns,
        dtype=float64,
    )


def dataframe_lr2sr(lr: DataFrame, /) -> DataFrame:
    return DataFrame(
        data=matrix_lr2sr(lr.values),
        index=lr.index,
        columns=lr.columns,
        dtype=float64,
    )


def dataframe_lr2crdd(lr: DataFrame, /) -> dict[str, DataFrame]:
    lr0_cr, lr0_dd = matrix_lr2crdd(lr.values)
    return {
        "LR_CR": DataFrame(
            data=lr0_cr, index=lr.index, columns=lr.columns, dtype=float64
        ),
        "LR_DD": DataFrame(
            data=lr0_dd, index=lr.index, columns=lr.columns, dtype=float64
        ),
    }


def dataframe_lr_metrics(
    lr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=matrix_lr_metrics(
            lr.values, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
        ),
        index=matrix_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=float64,
    )


def dataframe_lr_sr_metrics(
    lr: DataFrame,
    sr: DataFrame,
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> DataFrame:
    return DataFrame(
        data=matrix_lr_sr_metrics(
            lr.values,
            sr.values,
            ann_factor,
            sqrt_ann_factor=sqrt_ann_factor,
            mode=mode,
        ),
        index=matrix_lr_sr_metrics_meta(mode).keys(),
        columns=lr.columns,
        dtype=float64,
    )


def dataframe_portfolio_price(
    price: DataFrame,
    weights: DataFrame,
    /,
    fee_rate: float,
) -> Series:
    price = price[weights.columns]
    price = price.loc[weights.index]
    return Series(
        data=matrix_portfolio_price(price.values, weights.values, fee_rate),
        index=price.index,
        dtype=float64,
    )


def dataframe_risk_std_lr_cc(
    price: DataFrame, /, window: int, *, col_c: str
) -> Series:
    c = price[col_c]
    c1 = series_shift(c, 1)
    lr = _log(c / c1)
    return series_rolling_std(lr, window)


def dataframe_risk_std_lr_oc(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    c1 = series_shift(c, 1)
    lr = _log(o / c1)
    return series_rolling_std(lr, window)


def dataframe_risk_std_lr_co(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    lr = _log(c / o)
    return series_rolling_std(lr, window)


def dataframe_risk_std_sr_cc(
    price: DataFrame, /, window: int, *, col_c: str
) -> Series:
    c = price[col_c]
    c1 = series_shift(c, 1)
    sr = c / c1 - 1
    return series_rolling_std(sr, window)


def dataframe_risk_std_sr_oc(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    c1 = series_shift(c, 1)
    sr = o / c1 - 1
    return series_rolling_std(sr, window)


def dataframe_risk_std_sr_co(
    price: DataFrame, /, window: int, *, col_o: str, col_c: str
) -> Series:
    o = price[col_o]
    c = price[col_c]
    sr = c / o - 1
    return series_rolling_std(sr, window)


def dataframe_risk_atr(
    price: DataFrame, /, window: int, *, col_h: str, col_l: str, col_c: str
) -> Series:
    h = price[col_h]
    l = price[col_l]
    c = price[col_c]
    c1 = series_shift(c, 1)
    hl = h - l
    hc = _abs(h - c1)
    lc = _abs(l - c1)
    tr = DataFrame({"hl": hl, "hc": hc, "lc": lc}).max(axis=1, skipna=False)
    return series_rolling_mean(tr, window)


def dataframe_risk_ahl(
    price: DataFrame, /, window: int, *, col_h: str, col_l: str
) -> Series:
    h = price[col_h]
    l = price[col_l]
    hl = h - l
    return series_rolling_mean(hl, window)


def dataframe_risk_anr(
    price: DataFrame,
    /,
    window: int,
    *,
    col_o: str,
    col_h: str,
    col_l: str,
    col_c: str,
) -> Series:
    o = price[col_o]
    h = price[col_h]
    l = price[col_l]
    c = price[col_c]
    hl = h - l
    aco = _abs(c - o)
    noise = 1 - aco / hl
    return series_rolling_mean(noise, window)
