from typing import Callable, Optional

from bottleneck import allnan as _allnan
from bottleneck import move_argmax as _move_argmax
from bottleneck import move_argmin as _move_argmin
from bottleneck import move_max as _move_max
from bottleneck import move_mean as _move_mean
from bottleneck import move_median as _move_median
from bottleneck import move_min as _move_min
from bottleneck import move_rank as _move_rank
from bottleneck import move_std as _move_std
from bottleneck import move_sum as _move_sum
from bottleneck import move_var as _move_var
from bottleneck import nanmax as _nanmax
from bottleneck import nanmean as _nanmean
from bottleneck import nanmedian as _nanmedian
from bottleneck import nanmin as _nanmin
from bottleneck import nanstd as _nanstd
from bottleneck import nansum as _nansum
from numpy import abs as _abs
from numpy import asarray as _asarray
from numpy import concatenate as _concatenate
from numpy import float64
from numpy import full as _full
from numpy import isnan as _isnan
from numpy import log as _log
from numpy import maximum as _maximum
from numpy import nan as _nan
from numpy import nancumsum as _nancumsum
from numpy import nanpercentile as _nanpercentile
from numpy import sqrt as _sqrt
from numpy import zeros as _zeros
from numpy.lib.stride_tricks import sliding_window_view as _sliding_window_view
from numpy.typing import NDArray
from scipy.stats import kurtosis as _kurtosis
from scipy.stats import skew as _skew

from tompy.vector import (
    __lr2sr,
    __sr2lr,
    vector_describe_meta,
    vector_lr_sr_metrics,
    vector_lr_sr_metrics_meta,
    vector_zeros,
)


def matrix_validate(m: NDArray[float64], /) -> tuple[int, int]:
    assert m.ndim == 2
    return m.shape  # type: ignore


def matrix_validate_rolling(
    m: NDArray[float64], /, window: int
) -> tuple[int, int]:
    assert window >= 0
    return matrix_validate(m)


def matrix_zeros(nr: int, nc: int, /) -> NDArray[float64]:
    return _zeros((nr, nc), dtype=float64)


def matrix_nans(nr: int, nc: int, /) -> NDArray[float64]:
    return _full((nr, nc), _nan, dtype=float64)


def matrix_describe_meta() -> dict[str, str]:
    return vector_describe_meta()


def matrix_describe(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(len(matrix_describe_meta()), nc)
    return _asarray(
        [
            matrix_nancnt(m),
            matrix_nanmean(m),
            matrix_nanstd(m),
            matrix_nanskew(m),
            matrix_nankurt(m),
            matrix_nanpctl(m, 0),
            matrix_nanpctl(m, 25),
            matrix_nanpctl(m, 50),
            matrix_nanpctl(m, 75),
            matrix_nanpctl(m, 100),
        ],
        dtype=float64,
    )


def matrix_shift(m: NDArray[float64], /, lag: int) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0 or lag == 0:
        return m
    if lag > 0:
        if lag >= nr:
            return matrix_nans(nr, nc)
        return _concatenate((matrix_nans(lag, nc), m[:-lag, :]))
    lag = -lag
    if lag >= nr:
        return matrix_nans(nr, nc)
    return _concatenate((m[lag:, :], matrix_nans(lag, nc)))


def matrix_nancnt(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return nr - _nansum(_isnan(m), axis=0)  # type: ignore


def matrix_nanmean(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return _nanmean(m, axis=0)  # type: ignore


def matrix_nanstd(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    return _nanstd(m, axis=0, ddof=1)  # type: ignore


def matrix_nanskew(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if _allnan(m):
        return matrix_nans(1, nc)
    return _skew(m, axis=0, bias=False, nan_policy="omit")  # type: ignore


def matrix_nankurt(m: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if _allnan(m):
        return matrix_nans(1, nc)
    return _kurtosis(m, axis=0, fisher=True, bias=False, nan_policy="omit")  # type: ignore


def matrix_nanpctl(m: NDArray[float64], /, q: int) -> NDArray[float64]:
    nr, nc = matrix_validate(m)
    if nr <= 0:
        return matrix_nans(0, nc)
    if q <= 0:
        return _nanmin(m, axis=0)  # type: ignore
    if q >= 100:
        return _nanmax(m, axis=0)  # type: ignore
    if q == 50:
        return _nanmedian(m, axis=0)  # type: ignore
    if _allnan(m):
        return matrix_nans(1, nc)
    return _nanpercentile(m, q, axis=0)  # type: ignore


def matrix_rolling_sum(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_sum(m, window, axis=0)  # type: ignore


def matrix_rolling_mean(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_mean(m, window, axis=0)  # type: ignore


def matrix_rolling_std(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_std(m, window, axis=0, ddof=1)  # type: ignore


def matrix_rolling_var(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_var(m, window, axis=0, ddof=1)  # type: ignore


def matrix_rolling_min(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_min(m, window, axis=0)  # type: ignore


def matrix_rolling_max(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_max(m, window, axis=0)  # type: ignore


def matrix_rolling_argmin(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_argmin(m, window, axis=0)  # type: ignore


def matrix_rolling_argmax(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_argmax(m, window, axis=0)  # type: ignore


def matrix_rolling_median(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_median(m, window, axis=0)  # type: ignore


def matrix_rolling_rank(
    m: NDArray[float64], /, window: int
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nc)
    return _move_rank(m, window, axis=0)  # type: ignore


def matrix_rolling_apply(
    m: NDArray[float64],
    /,
    window: int,
    func: Callable[[NDArray[float64]], NDArray[float64]],
    nfout: int,
) -> NDArray[float64]:
    nr, nc = matrix_validate_rolling(m, window)
    if window == 0 or window > nr:
        return matrix_nans(nr, nfout)
    swv = _sliding_window_view(m, (window, nc)).reshape((-1, window, nc))
    ret = _asarray([func(sw) for sw in swv], dtype=float64)
    assert ret.ndim == 2 and ret.shape[1] == nfout
    if window > 1:
        ret = _concatenate((matrix_nans(window - 1, nfout), ret))
    return ret


def matrix_price2relative(price: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    rel = price[1:, :] / price[:-1, :]
    return _concatenate((matrix_nans(1, nc), rel))


def matrix_price2sr(price: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    sr = price[1:, :] / price[:-1, :] - 1
    return _concatenate((matrix_nans(1, nc), sr))


def matrix_price2lr(price: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(price)
    if nr <= 1 or nc <= 0:
        return matrix_nans(nr, nc)
    lr = _log(price[1:, :] / price[:-1, :])
    return _concatenate((matrix_nans(1, nc), lr))


def matrix_price_metrics(
    price: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    rel = matrix_price2relative(price)
    sr = rel - 1
    lr = _log(rel)
    return matrix_lr_sr_metrics(
        lr, sr, ann_factor, sqrt_ann_factor=sqrt_ann_factor, mode=mode
    )


def matrix_sr2lr(sr: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(sr)
    if nr <= 0:
        return matrix_nans(nr, nc)
    return __sr2lr(sr)


def matrix_sr_metrics(
    sr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return matrix_lr_sr_metrics(
        matrix_sr2lr(sr),
        sr,
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def matrix_lr2sr(lr: NDArray[float64], /) -> NDArray[float64]:
    nr, nc = matrix_validate(lr)
    if nr <= 0:
        return matrix_nans(nr, nc)
    return __lr2sr(lr)


def matrix_lr2crdd(
    lr: NDArray[float64], /
) -> tuple[NDArray[float64], NDArray[float64]]:
    nr, nc = matrix_validate(lr)
    if nr <= 0:
        return matrix_nans(nr, nc), matrix_nans(nr, nc)
    lr0_cr = _nancumsum(lr, axis=0)
    lr0_peak = _maximum.accumulate(lr0_cr, axis=0)
    lr0_dd = lr0_cr - lr0_peak
    return lr0_cr, lr0_dd


def matrix_lr_metrics(
    lr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    return matrix_lr_sr_metrics(
        lr,
        matrix_lr2sr(lr),
        ann_factor,
        sqrt_ann_factor=sqrt_ann_factor,
        mode=mode,
    )


def matrix_lr_sr_metrics_meta(mode: int, /) -> dict[str, str]:
    return vector_lr_sr_metrics_meta(mode)


def matrix_lr_sr_metrics(
    lr: NDArray[float64],
    sr: NDArray[float64],
    /,
    ann_factor: float,
    *,
    sqrt_ann_factor: Optional[float] = None,
    mode: int = 0,
) -> NDArray[float64]:
    nr_lr, nc_lr = matrix_validate(lr)
    nr_sr, nc_sr = matrix_validate(sr)
    assert nr_lr == nr_sr and nc_lr == nc_sr
    if nr_lr <= 0:
        return matrix_nans(len(matrix_lr_sr_metrics_meta(mode)), nc_lr)
    if sqrt_ann_factor is None:
        sqrt_ann_factor = _sqrt(ann_factor)
    return _asarray(
        [
            vector_lr_sr_metrics(
                lr[:, i],
                sr[:, i],
                ann_factor,
                sqrt_ann_factor=sqrt_ann_factor,
                mode=mode,
            )
            for i in range(nc_lr)
        ],
        dtype=float64,
    ).T


def matrix_portfolio_price(
    price: NDArray[float64], weight: NDArray[float64], /, fee_rate: float
) -> NDArray[float64]:
    price_shape = matrix_validate(price)
    weight_shape = matrix_validate(weight)
    assert price_shape == weight_shape
    nr, nc = weight_shape
    nav = 1
    position_quantity = vector_zeros(nc)
    cash = 1
    navs = vector_zeros(nr)
    for i in range(nr):
        pi = price[i]
        wi = weight[i]
        nav = cash + _nansum(position_quantity * pi)
        if not _allnan(wi):
            target_position_dollar = nav * wi
            target_position_quantity = target_position_dollar / pi
            trading_position_quantity = (
                target_position_quantity - position_quantity
            )
            trading_position_dollar = trading_position_quantity * pi
            trading_fee = _nansum(_abs(trading_position_dollar)) * fee_rate
            nav = (
                cash
                - _nansum(trading_position_dollar)
                + _nansum(target_position_dollar)
                - trading_fee
            )
            position_dollar = nav * wi
            position_quantity = position_dollar / pi
            cash = nav - _nansum(position_dollar)
        navs[i] = nav
    return navs
