from abc import ABC, abstractmethod
from typing import Optional

from numpy import full as _full
from numpy import nan as _nan
from numpy import tile as _tile
from pandas import DataFrame, Series

from tompy.dataframe import dataframe_portfolio_price, dataframe_shift


def rebalancing_weight(
    weights: DataFrame,
    /,
    rebalancing: str,
    lag: int,
) -> DataFrame:
    if rebalancing == "W":
        idx = DataFrame(index=weights.index)
        idx["W"] = idx.index.dayofweek
        idx["NA"] = idx["W"].diff(-1).fillna(0) <= 0
        weights = weights.copy()
        weights.loc[idx["NA"]] = _nan
    elif rebalancing == "M":
        idx = DataFrame(index=weights.index)
        idx["M"] = idx.index.month
        idx["NA"] = idx["M"].diff(-1).fillna(0) == 0
        weights = weights.copy()
        weights.loc[idx["NA"]] = _nan
    return dataframe_shift(weights, lag)


def crp_ew(price: DataFrame, tickers: list[str], /) -> DataFrame:
    price = price[tickers].dropna()
    nr = price.shape[0]
    nc = len(tickers)
    ew = 1.0 / nc
    m = _full((nr, nc), ew)
    return DataFrame(data=m, index=price.index, columns=tickers)


def crp(price: DataFrame, ticker_weights: dict[str, float], /) -> DataFrame:
    tickers = list(ticker_weights.keys())
    weights = list(ticker_weights.values())
    price = price[tickers].dropna()
    nr = price.shape[0]
    m = _tile(weights, (nr, 1))
    return DataFrame(data=m, index=price.index, columns=tickers)


class Base(ABC):
    def __init__(
        self,
        price: DataFrame,
        /,
    ) -> None:
        self.__price = price

    @property
    def price(self, /) -> DataFrame:
        return self.__price

    @abstractmethod
    def portfolio_weight(self, /) -> DataFrame:
        raise NotImplementedError

    def rebalancing_weight(
        self,
        /,
        *,
        weights: Optional[DataFrame] = None,
        rebalancing: str = "D",
        lag: int = 1,
    ) -> DataFrame:
        if weights is None:
            weights = self.portfolio_weight()
        return rebalancing_weight(weights, rebalancing, lag)

    def portfolio_price(
        self, weights: DataFrame, /, fee_rate: float
    ) -> Series:
        return dataframe_portfolio_price(self.price, weights, fee_rate)


class CRP_EW(Base):
    def __init__(
        self,
        price: DataFrame,
        tickers: list[str],
        /,
    ) -> None:
        super().__init__(price)
        self.__tickers = list(dict.fromkeys(tickers))

    @property
    def tickers(self, /) -> list[str]:
        return self.__tickers

    def portfolio_weight(self, /) -> DataFrame:
        return crp_ew(self.price, self.tickers)


class CRP(Base):
    def __init__(
        self,
        price: DataFrame,
        ticker_weights: dict[str, float],
        /,
    ) -> None:
        super().__init__(price)
        self.__ticker_weights = ticker_weights

    @property
    def ticker_weights(self, /) -> dict[str, float]:
        return self.__ticker_weights

    def portfolio_weight(self, /) -> DataFrame:
        return crp(self.price, self.ticker_weights)
